@@ -1,8 +1,8 @@
 /* Routines invoked by a function key
    They normally operate on the current panel.
    
-   Copyright (C) 1994, 1995 Miguel de Icaza
-   Copyright (C) 1994, 1995 Janne Kukonlehto
+   Copyright (C) 1994, 1995, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
+   2005, 2006, 2007 Free Software Foundation, Inc.
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -34,6 +34,8 @@
 #endif
 #include <unistd.h>
 
+#include <mhl/string.h>
+
 #include "global.h"
 #include "cmd.h"		/* Our definitions */
 #include "fileopctx.h"		/* file_op_context_new() */
@@ -61,6 +63,7 @@
 #include "setup.h"		/* save_setup() */
 #include "profile.h"		/* PROFILE_NAME */
 #include "execute.h"		/* toggle_panels() */
+#include "history.h"
 
 #ifndef MAP_FILE
 #   define MAP_FILE 0
@@ -197,8 +200,8 @@ do_view_cmd (int normal)
 	if (!do_cd (selection (current_panel)->fname, cd_exact))
 	    message (1, MSG_ERROR, _("Cannot change directory"));
 
+	repaint_screen();
 	return;
-
     }
 
     file_idx = current_panel->selected;
@@ -212,6 +215,8 @@ do_view_cmd (int normal)
 	    break;
 	file_idx = scan_for_file (current_panel, file_idx, dir);
     }
+
+    repaint_screen();
 }
 
 /* Run user's preferred viewer on the currently selected file */
@@ -229,7 +234,7 @@ view_file_cmd (void)
 
     filename =
 	input_expand_dialog (_(" View file "), _(" Filename:"),
-			     selection (current_panel)->fname);
+			     MC_HISTORY_FM_VIEW_FILE, selection (current_panel)->fname);
     if (!filename)
 	return;
 
@@ -252,6 +257,7 @@ filtered_view_cmd (void)
     command =
 	input_dialog (_(" Filtered view "),
 		      _(" Filter command and arguments:"),
+		      MC_HISTORY_FM_FILTERED_VIEW,
 		      selection (current_panel)->fname);
     if (!command)
 	return;
@@ -350,14 +356,15 @@ mkdir_cmd (void)
 
     dir =
 	input_expand_dialog (_("Create a new Directory"),
-			     _(" Enter directory name:"), "");
+			     _(" Enter directory name:"), 
+			     MC_HISTORY_FM_MKDIR, "");
     if (!dir)
 	return;
 
     if (dir[0] == '/' || dir[0] == '~')
 	absdir = g_strdup (dir);
     else
-	absdir = concat_dir_and_file (current_panel->cwd, dir);
+	absdir = mhl_str_dir_plus_file (current_panel->cwd, dir);
 
     save_cwds_stat ();
     if (my_mkdir (absdir, 0777) == 0) {
@@ -422,7 +429,7 @@ set_panel_filter (WPanel *p)
 	
     reg_exp = input_dialog_help (_(" Filter "),
 				 _(" Set expression for filtering filenames"),
-				 "[Filter...]", x);
+				 "[Filter...]", MC_HISTORY_FM_PANEL_FILTER, x);
     if (!reg_exp)
 	return;
     set_panel_filter_to (p, reg_exp);
@@ -468,14 +475,14 @@ void reverse_selection_cmd (void)
 }
 
 static void
-select_unselect_cmd (const char *title, int cmd)
+select_unselect_cmd (const char *title, const char *history_name, int cmd)
 {
     char *reg_exp, *reg_exp_t;
     int i;
     int c;
     int dirflag = 0;
 
-    reg_exp = input_dialog (title, "", easy_patterns ? "*" : ".");
+    reg_exp = input_dialog (title, "", history_name, easy_patterns ? "*" : ".");
     if (!reg_exp)
 	return;
     if (!*reg_exp) {
@@ -521,12 +528,12 @@ select_unselect_cmd (const char *title, int cmd)
 
 void select_cmd (void)
 {
-    select_unselect_cmd (_(" Select "), 1);
+    select_unselect_cmd (_(" Select "), ":select_cmd: Select ", 1);
 }
 
 void unselect_cmd (void)
 {
-    select_unselect_cmd (_(" Unselect "), 0);
+    select_unselect_cmd (_(" Unselect "), ":unselect_cmd: Unselect ", 0);
 }
 
 /* Check if the file exists */
@@ -563,10 +570,10 @@ void ext_cmd (void)
 			    _(" Which extension file you want to edit? "), 0, 2,
 			    _("&User"), _("&System Wide"));
     }
-    extdir = concat_dir_and_file (mc_home, MC_LIB_EXT);
+    extdir = mhl_str_dir_plus_file (mc_home, MC_LIB_EXT);
 
     if (dir == 0){
-	buffer = concat_dir_and_file (home_dir, MC_USER_EXT);
+	buffer = mhl_str_dir_plus_file (home_dir, MC_USER_EXT);
 	check_for_default (extdir, buffer);
 	do_edit (buffer);
 	g_free (buffer);
@@ -590,10 +597,10 @@ menu_edit_cmd (int where)
 	_(" Menu edit "),
 	_(" Which menu file do you want to edit? "), 
 	0, geteuid() ? 2 : 3,
-	_("&Local"), _("&Home"), _("&System Wide")
+	_("&Local"), _("&User"), _("&System Wide")
     );
 
-    menufile = concat_dir_and_file (mc_home, where ? CEDIT_GLOBAL_MENU : MC_GLOBAL_MENU);
+    menufile = mhl_str_dir_plus_file (mc_home, where ? CEDIT_GLOBAL_MENU : MC_GLOBAL_MENU);
 
     switch (dir) {
 	case 0:
@@ -602,12 +609,12 @@ menu_edit_cmd (int where)
 	    break;
 
 	case 1:
-	    buffer = concat_dir_and_file (home_dir, where ? CEDIT_HOME_MENU : MC_HOME_MENU);
+	    buffer = mhl_str_dir_plus_file (home_dir, where ? CEDIT_HOME_MENU : MC_HOME_MENU);
 	    check_for_default (menufile, buffer);
 	    break;
 	
 	case 2:
-	    buffer = concat_dir_and_file (mc_home, where ? CEDIT_GLOBAL_MENU : MC_GLOBAL_MENU);
+	    buffer = mhl_str_dir_plus_file (mc_home, where ? CEDIT_GLOBAL_MENU : MC_GLOBAL_MENU);
 	    break;
 
 	default:
@@ -666,10 +673,10 @@ edit_syntax_cmd (void)
 			  _(" Which syntax file you want to edit? "), 0, 2,
 			  _("&User"), _("&System Wide"));
     }
-    extdir = concat_dir_and_file (mc_home, "syntax" PATH_SEP_STR "Syntax");
+    extdir = mhl_str_dir_plus_file (mc_home, "syntax" PATH_SEP_STR "Syntax");
 
     if (dir == 0) {
-	buffer = concat_dir_and_file (home_dir, SYNTAX_FILE);
+	buffer = mhl_str_dir_plus_file (home_dir, SYNTAX_FILE);
 	check_for_default (extdir, buffer);
 	do_edit (buffer);
 	g_free (buffer);
@@ -800,8 +807,8 @@ compare_dir (WPanel *panel, WPanel *other, enum CompareMode mode)
 	    }
 
 	    /* Thorough compare on, do byte-by-byte comparison */
-	    src_name = concat_dir_and_file (panel->cwd, source->fname);
-	    dst_name = concat_dir_and_file (other->cwd, target->fname);
+	    src_name = mhl_str_dir_plus_file (panel->cwd, source->fname);
+	    dst_name = mhl_str_dir_plus_file (other->cwd, target->fname);
 	    if (compare_files (src_name, dst_name, source->st.st_size))
 		do_file_mark (panel, i, 1);
 	    g_free (src_name);
@@ -903,7 +910,7 @@ do_link (int symbolic_link, const char *fname)
 
     if (!symbolic_link) {
 	src = g_strdup_printf (_("Link %s to:"), name_trunc (fname, 46));
-	dest = input_expand_dialog (_(" Link "), src, "");
+	dest = input_expand_dialog (_(" Link "), src, MC_HISTORY_FM_LINK, "");
 	if (!dest || !*dest)
 	    goto cleanup;
 	save_cwds_stat ();
@@ -915,10 +922,10 @@ do_link (int symbolic_link, const char *fname)
 	char *d;
 
 	/* suggest the full path for symlink */
-	s = concat_dir_and_file (current_panel->cwd, fname);
+	s = mhl_str_dir_plus_file (current_panel->cwd, fname);
 
 	if (get_other_type () == view_listing) {
-	    d = concat_dir_and_file (other_panel->cwd, fname);
+	    d = mhl_str_dir_plus_file (other_panel->cwd, fname);
 	} else {
 	    d = g_strdup (fname);
 	}
@@ -972,7 +979,7 @@ void edit_symlink_cmd (void)
 	i = readlink (p, buffer, MC_MAXPATHLEN - 1);
 	if (i > 0) {
 	    buffer [i] = 0;
-	    dest = input_expand_dialog (_(" Edit symlink "), q, buffer);
+	    dest = input_expand_dialog (_(" Edit symlink "), q, MC_HISTORY_FM_EDIT_LINK, buffer);
 	    if (dest) {
 		if (*dest && strcmp (buffer, dest)) {
 		    save_cwds_stat ();
@@ -1085,7 +1092,8 @@ get_random_hint (int force)
 
 #if defined(USE_NETCODE) || defined(USE_EXT2FSLIB)
 static void
-nice_cd (const char *text, const char *xtext, const char *help, const char *prefix, int to_home)
+nice_cd (const char *text, const char *xtext, const char *help,
+	 const char *history_name, const char *prefix, int to_home)
 {
     char *machine;
     char *cd_path;
@@ -1093,9 +1101,7 @@ nice_cd (const char *text, const char *xtext, const char *help, const char *pref
     if (!SELECTED_IS_PANEL)
 	return;
 
-    machine = input_dialog_help (text,
-				 xtext,
-				 help, "");
+    machine = input_dialog_help (text, xtext, help, history_name, "");
     if (!machine)
 	return;
 
@@ -1125,27 +1131,30 @@ static const char *machine_str = N_(" Enter machine name (F1 for details): ");
 void netlink_cmd (void)
 {
     nice_cd (_(" Link to a remote machine "), _(machine_str),
-	     "[Network File System]", "/#mc:", 1);
+	     "[Network File System]", ":netlink_cmd: Link to a remote ",
+	     "/#mc:", 1);
 }
 #endif /* WITH_MCFS */
 
 void ftplink_cmd (void)
 {
     nice_cd (_(" FTP to machine "), _(machine_str),
-	     "[FTP File System]", "/#ftp:", 1);
+	     "[FTP File System]", ":ftplink_cmd: FTP to machine ", "/#ftp:", 1);
 }
 
 void fishlink_cmd (void)
 {
     nice_cd (_(" Shell link to machine "), _(machine_str),
-	     "[FIle transfer over SHell filesystem]", "/#sh:", 1);
+	     "[FIle transfer over SHell filesystem]", ":fishlink_cmd: Shell link to machine ",
+	     "/#sh:", 1);
 }
 
 #ifdef WITH_SMBFS
 void smblink_cmd (void)
 {
     nice_cd (_(" SMB link to machine "), _(machine_str),
-	     "[SMB File System]", "/#smb:", 0);
+	     "[SMB File System]", ":smblink_cmd: SMB link to machine ",
+	     "/#smb:", 0);
 }
 #endif /* WITH_SMBFS */
 #endif /* USE_NETCODE */
@@ -1156,7 +1165,8 @@ void undelete_cmd (void)
     nice_cd (_(" Undelete files on an ext2 file system "),
 	     _(" Enter device (without /dev/) to undelete\n "
 	       "  files on: (F1 for details)"),
-	     "[Undelete File System]", "/#undel:", 0);
+	     "[Undelete File System]", ":undelete_cmd: Undel on ext2 fs ",
+	     "/#undel:", 0);
 }
 #endif /* USE_EXT2FSLIB */
 
@@ -1216,20 +1226,17 @@ dirsizes_cmd (void)
 	}
 	
     recalculate_panel_summary (panel);
+    panel_re_sort (panel);
     panel->dirty = 1;
 }
 
 void
 save_setup_cmd (void)
 {
-    char *str;
-    
     save_setup ();
     sync_profiles ();
-    str = g_strconcat ( _(" Setup saved to ~/"), PROFILE_NAME, (char *) NULL);
     
-    message (0, _(" Setup "), str);
-    g_free (str);
+    message (0, _(" Setup "), _(" Setup saved to ~/%s"), PROFILE_NAME);
 }
 
 static void
