@@ -1,5 +1,6 @@
 /* Client interface for General purpose Linux console save/restore server
-   Copyright (C) 1994 Janne Kukonlehto <jtklehto@stekt.oulu.fi>
+   Copyright (C) 1994, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
+   2006, 2007 Free Software Foundation, Inc. 
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,6 +28,8 @@
 #endif
 #include <unistd.h>
 
+#include <mhl/string.h>
+
 #include "global.h"
 #include "tty.h"
 #include "cons.saver.h"
@@ -142,7 +145,7 @@ handle_console_linux (unsigned char action)
 	    open ("/dev/null", O_WRONLY);
 	    if (tty_name) {
 		/* Exec the console save/restore handler */
-		mc_conssaver = concat_dir_and_file (SAVERDIR, "cons.saver");
+		mc_conssaver = mhl_str_dir_plus_file (SAVERDIR, "cons.saver");
 		execl (mc_conssaver, "cons.saver", tty_name, (char *) NULL);
 	    }
 	    /* Console is not a tty or execl() failed */
