@@ -1,5 +1,6 @@
 /* Find file command for the Midnight Commander
-   Copyright (C) The Free Software Foundation
+   Copyright (C) 1995, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
+   2006, 2007 Free Software Foundation, Inc.
    Written 1995 by Miguel de Icaza
 
    Complete rewrote.
@@ -19,10 +20,14 @@
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 #include <config.h>
-#include <string.h>
-#include <stdio.h>
-#include <sys/stat.h>
+
 #include <ctype.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/stat.h>
+
+#include <mhl/string.h>
 
 #include "global.h"
 #include "tty.h"
@@ -44,7 +49,7 @@
 #include "key.h"
 
 /* Size of the find parameters window */
-#define FIND_Y 14
+#define FIND_Y 15
 static int FIND_X = 50;
 
 /* Size of the find window */
@@ -75,10 +80,13 @@ static WInput *in_start;	/* Start path */
 static WInput *in_name;		/* Pattern to search */
 static WInput *in_with;		/* Text inside filename */
 static WCheck *case_sense;	/* "case sensitive" checkbox */
+static WCheck *find_regex_cbox;	/* [x] find regular expression */
 
 static int running = 0;		/* nice flag */
 static char *find_pattern;	/* Pattern to search */
-static char *content_pattern;	/* pattern to search inside files */
+static char *content_pattern;	/* pattern to search inside files; if
+				   find_regex_flag is true, it contains the
+				   regex pattern, else the search string. */
 static int count;		/* Number of files displayed */
 static int matches;		/* Number of matches */
 static int is_start;		/* Status of the start/stop toggle button */
@@ -135,6 +143,8 @@ static void get_list_info (char **file, char **dir) {
 static regex_t *r; /* Pointer to compiled content_pattern */
  
 static int case_sensitive = 1;
+static gboolean find_regex_flag = TRUE;
+static int find_recursively = 1;
 
 /*
  * Callback for the parameter dialog.
@@ -147,7 +157,8 @@ find_parm_callback (struct Dlg_head *h, dlg_msg_t msg, int parm)
 
     switch (msg) {
     case DLG_VALIDATE:
-	if ((h->ret_value != B_ENTER) || !in_with->buffer[0])
+	if ((h->ret_value != B_ENTER) || !in_with->buffer[0]
+	    || !(find_regex_cbox->state & C_BOOL))
 	    return MSG_HANDLED;
 
 	flags = REG_EXTENDED | REG_NOSUB;
@@ -157,7 +168,7 @@ find_parm_callback (struct Dlg_head *h, dlg_msg_t msg, int parm)
 
 	if (regcomp (r, in_with->buffer, flags)) {
 	    message (1, MSG_ERROR, _("  Malformed regular expression  "));
-	    dlg_select_widget (h, in_with);
+	    dlg_select_widget (in_with);
 	    h->running = 1;	/* Don't stop the dialog */
 	}
 	return MSG_HANDLED;
@@ -186,6 +197,9 @@ find_parameters (char **start_dir, char **pattern, char **content)
     int return_value;
     char *temp_dir;
     static const char *case_label = N_("case &Sensitive");
+    static const char *recurs_label = N_("&Find recursively");
+
+    WCheck *recursively_cbox;
 
     static char *in_contents = NULL;
     static char *in_start_dir = NULL;
@@ -228,6 +242,7 @@ find_parameters (char **start_dir, char **pattern, char **content)
 
 	i18n_flag = 1;
 	case_label = _(case_label);
+	recurs_label = _(recurs_label);
     }
 #endif				/* ENABLE_NLS */
 
@@ -245,32 +260,39 @@ find_parameters (char **start_dir, char **pattern, char **content)
 		    DLG_CENTER | DLG_REVERSE);
 
     add_widget (find_dlg,
-		button_new (11, b2, B_CANCEL, NORMAL_BUTTON, buts[2], 0));
+		button_new (12, b2, B_CANCEL, NORMAL_BUTTON, buts[2], 0));
     add_widget (find_dlg,
-		button_new (11, b1, B_TREE, NORMAL_BUTTON, buts[1], 0));
+		button_new (12, b1, B_TREE, NORMAL_BUTTON, buts[1], 0));
     add_widget (find_dlg,
-		button_new (11, b0, B_ENTER, DEFPUSH_BUTTON, buts[0], 0));
+		button_new (12, b0, B_ENTER, DEFPUSH_BUTTON, buts[0], 0));
 
-    case_sense = check_new (9, 3, case_sensitive, case_label);
+    recursively_cbox =
+	check_new (6, istart, find_recursively, recurs_label);
+ 
+    find_regex_cbox = check_new (10, istart, find_regex_flag, _("&Regular expression"));
+    add_widget (find_dlg, find_regex_cbox);
+
+    case_sense = check_new (9, istart, case_sensitive, case_label);
     add_widget (find_dlg, case_sense);
 
     in_with =
-	input_new (7, istart, INPUT_COLOR, ilen, in_contents, "content");
+	input_new (8, istart, INPUT_COLOR, ilen, in_contents, "content", INPUT_COMPLETE_DEFAULT);
     add_widget (find_dlg, in_with);
 
+    add_widget (find_dlg, recursively_cbox);
     in_name =
-	input_new (5, istart, INPUT_COLOR, ilen, in_start_name, "name");
+	input_new (5, istart, INPUT_COLOR, ilen, in_start_name, "name", INPUT_COMPLETE_DEFAULT);
     add_widget (find_dlg, in_name);
 
     in_start =
-	input_new (3, istart, INPUT_COLOR, ilen, in_start_dir, "start");
+	input_new (3, istart, INPUT_COLOR, ilen, in_start_dir, "start", INPUT_COMPLETE_DEFAULT);
     add_widget (find_dlg, in_start);
 
-    add_widget (find_dlg, label_new (7, 3, labs[2]));
+    add_widget (find_dlg, label_new (8, 3, labs[2]));
     add_widget (find_dlg, label_new (5, 3, labs[1]));
     add_widget (find_dlg, label_new (3, 3, labs[0]));
 
-    dlg_select_widget (find_dlg, in_name);
+    dlg_select_widget (in_name);
 
     run_dlg (find_dlg);
 
@@ -282,6 +304,8 @@ find_parameters (char **start_dir, char **pattern, char **content)
     case B_TREE:
 	temp_dir = g_strdup (in_start->buffer);
 	case_sensitive = case_sense->state & C_BOOL;
+	find_regex_flag = find_regex_cbox->state & C_BOOL;
+ 	find_recursively = recursively_cbox->state & C_BOOL;
 	destroy_dlg (find_dlg);
 	g_free (in_start_dir);
 	if (strcmp (temp_dir, ".") == 0) {
@@ -308,6 +332,8 @@ find_parameters (char **start_dir, char **pattern, char **content)
 	}
 
 	case_sensitive = case_sense->state & C_BOOL;
+	find_regex_flag = find_regex_cbox->state & C_BOOL;
+ 	find_recursively = recursively_cbox->state & C_BOOL;
 	return_value = 1;
 	*start_dir = g_strdup (in_start->buffer);
 	*pattern = g_strdup (in_name->buffer);
@@ -329,7 +355,7 @@ push_directory (const char *dir)
     dir_stack *new;
 
     new = g_new (dir_stack, 1);
-    new->name = concat_dir_and_file (dir, "");
+    new->name = mhl_str_dir_plus_file (dir, NULL);
     new->prev = dir_stack_base;
     dir_stack_base = new;
 }
@@ -380,6 +406,8 @@ find_add_match (Dlg_head *h, const char *dir, const char *file)
 {
     int p = ++matches & 7;
 
+    (void) h;
+
     insert_file (dir, file);
 
     /* Scroll nicely */
@@ -489,7 +517,7 @@ search_content (Dlg_head *h, const char *directory, const char *filename)
     int file_fd;
     int ret_val = 0;
 
-    fname = concat_dir_and_file (directory, filename);
+    fname = mhl_str_dir_plus_file (directory, filename);
 
     if (mc_stat (fname, &s) != 0 || !S_ISREG (s.st_mode)){
 	g_free (fname);
@@ -517,6 +545,8 @@ search_content (Dlg_head *h, const char *directory, const char *filename)
 	int has_newline;
 	char *p;
 	int found = 0;
+	typedef const char * (*search_fn) (const char *, const char *);
+	search_fn search_func;
 
 	if (resuming) {
 	    /* We've been previously suspended, start from the previous position */
@@ -525,14 +555,24 @@ search_content (Dlg_head *h, const char *directory, const char *filename)
 	    pos = last_pos;
 	}
 
+	search_func = (case_sensitive) ? cstrstr : cstrcasestr;
+	
 	while ((p = get_line_at (file_fd, buffer, &pos, &n_read, sizeof (buffer), &has_newline)) && (ret_val == 0)){
 	    if (found == 0){	/* Search in binary line once */
+	    	if (find_regex_flag) {
 		if (regexec (r, p, 1, 0, 0) == 0){
 		    g_free (p);
 		    p = g_strdup_printf ("%d:%s", line, filename);
 		    find_add_match (h, directory, p);
 		    found = 1;
 		}
+	    	} else {
+	    	    if (search_func (p, content_pattern) != NULL) {
+	    	    	char *match = g_strdup_printf("%d:%s", line, filename);
+			find_add_match (h, directory, match);
+			found = TRUE;
+	    	    }
+	    	}
 	    }
 	    if (has_newline){
 		line++;
@@ -652,8 +692,8 @@ do_search (struct Dlg_head *h)
 	return 1;
     }
 
-    if (subdirs_left && directory) { /* Can directory be NULL ? */
-	char *tmp_name = concat_dir_and_file (directory, dp->d_name);
+    if (subdirs_left && find_recursively && directory) { /* Can directory be NULL ? */
+	char *tmp_name = mhl_str_dir_plus_file (directory, dp->d_name);
 	if (!mc_lstat (tmp_name, &tmp_stat)
 	    && S_ISDIR (tmp_stat.st_mode)) {
 	    push_directory (tmp_name);
@@ -681,7 +721,7 @@ do_search (struct Dlg_head *h)
 
 	if (verbose){
 	    pos = (pos + 1) % 4;
-	    attrset (NORMALC);
+	    attrset (DLG_NORMALC (h));
 	    dlg_move (h, FIND2_Y-6, FIND2_X - 4);
 	    addch (rotating_dash [pos]);
 	    mc_refresh ();
@@ -706,6 +746,17 @@ init_find_vars (void)
 	g_free (dir);
 }
 
+static char *
+make_fullname (const char *dirname, const char *filename)
+{
+
+    if (strcmp(dirname, ".") == 0 || strcmp(dirname, "."PATH_SEP_STR) == 0)
+	return g_strdup (filename);
+    if (strncmp(dirname, "."PATH_SEP_STR, 2) == 0)
+	return mhl_str_dir_plus_file (dirname + 2, filename);
+    return mhl_str_dir_plus_file (dirname, filename);
+}
+
 static void
 find_do_view_edit (int unparsed_view, int edit, char *dir, char *file)
 {
@@ -720,13 +771,8 @@ find_do_view_edit (int unparsed_view, int edit, char *dir, char *file)
 	 filename = file + 4;
 	 line = 0;
     }
-    if (dir [0] == '.' && dir [1] == 0)
-	 fullname = g_strdup (filename);
-    else if (dir [0] == '.' && dir [1] == PATH_SEP)
-	 fullname = concat_dir_and_file (dir+2, filename);
-    else
-	 fullname = concat_dir_and_file (dir, filename);
 
+    fullname = make_fullname (dir, filename);
     if (edit)
 	do_edit_at_line (fullname, line);
     else
@@ -779,6 +825,8 @@ find_callback (struct Dlg_head *h, dlg_msg_t msg, int parm)
 static int
 start_stop (int button)
 {
+    (void) button;
+
     running = is_start;
     set_idle_proc (find_dlg, running);
     is_start = !is_start;
@@ -793,6 +841,8 @@ start_stop (int button)
 static int
 find_do_view_file (int button)
 {
+    (void) button;
+
     view_edit_currently_selected_file (0, 0);
     return 0;
 }
@@ -801,6 +851,8 @@ find_do_view_file (int button)
 static int
 find_do_edit_file (int button)
 {
+    (void) button;
+
     view_edit_currently_selected_file (0, 1);
     return 0;
 }
@@ -941,7 +993,7 @@ find_file (char *start_dir, char *pattern, char *content, char **dirname,
 	struct stat st;
 	WLEntry *entry = find_list->list;
 	dir_list *list = &current_panel->dir;
-	char *dir, *name;
+	char *name;
 
 	for (i = 0; entry && i < find_list->count;
 	     entry = entry->next, i++) {
@@ -955,13 +1007,7 @@ find_file (char *start_dir, char *pattern, char *content, char **dirname,
 	    else
 		filename = entry->text + 4;
 
-	    dir = entry->data;
-	    if (dir[0] == '.' && dir[1] == 0)
-		name = g_strdup (filename);
-	    else if (dir[0] == '.' && dir[1] == PATH_SEP)
-		name = concat_dir_and_file (dir + 2, filename);
-	    else
-		name = concat_dir_and_file (dir, filename);
+	    name = make_fullname (entry->data, filename);
 	    status =
 		handle_path (list, name, &st, next_free, &link_to_dir,
 			     &stale_link);
@@ -1023,7 +1069,7 @@ find_file (char *start_dir, char *pattern, char *content, char **dirname,
 void
 do_find (void)
 {
-    char *start_dir, *pattern, *content;
+    char *start_dir = NULL, *pattern = NULL, *content = NULL;
     char *filename, *dirname;
     int  v, dir_and_file_set;
     regex_t rx; /* Compiled content_pattern to search inside files */
@@ -1035,7 +1081,7 @@ do_find (void)
 	v = find_file (start_dir, pattern, content, &dirname, &filename);
 	g_free (start_dir);
 	g_free (pattern);
-	if (r)
+	if (find_regex_flag && r)
 	    regfree (r);
 	
 	if (v == B_ENTER){
@@ -1064,6 +1110,7 @@ do_find (void)
 	    if (dir_and_file_set){
 	        try_to_select (current_panel, NULL);
 		panel_re_sort (current_panel);
+	        try_to_select (current_panel, NULL);
 	    }
 	    break;
 	}
