@@ -1,5 +1,6 @@
 /* Panel managing.
-   Copyright (C) 1994, 1995 Miguel de Icaza.
+   Copyright (C) 1994, 1995, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
+   2005, 2006, 2007 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,9 +25,10 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-
 #include <unistd.h>
 
+#include <mhl/string.h>
+
 #include "global.h"
 #include "tty.h"
 #include "dir.h"
@@ -708,7 +710,7 @@ display_mini_info (WPanel *panel)
 	char *link, link_target [MC_MAXPATHLEN];
 	int  len;
 
-	link = concat_dir_and_file (panel->cwd, panel->dir.list [panel->selected].fname);
+	link = mhl_str_dir_plus_file (panel->cwd, panel->dir.list [panel->selected].fname);
 	len = mc_readlink (link, link_target, MC_MAXPATHLEN - 1);
 	g_free (link);
 	if (len > 0){
@@ -787,14 +789,17 @@ show_dir (WPanel *panel)
 	attrset (REVERSE_COLOR);
 
     widget_move (&panel->widget, 0, 3);
+    addch (' ');
 
     tmp = g_malloc (panel->widget.cols + 1);
     tmp[panel->widget.cols] = '\0';
 
     trim (strip_home_and_password (panel->cwd), tmp,
-	 min (max (panel->widget.cols - 7, 0), panel->widget.cols) );
+	 max (panel->widget.cols - 9, 0));
     addstr (tmp);
     g_free (tmp);
+
+    addch (' ');
     widget_move (&panel->widget, 0, 1);
     addstr ("<");
     widget_move (&panel->widget, 0, panel->widget.cols - 2);
@@ -1050,7 +1055,8 @@ panel_new (const char *panel_name)
     /* Load the default format */
     panel->count =
 	do_load_dir (panel->cwd, &panel->dir, panel->sort_type,
-		     panel->reverse, panel->case_sensitive, panel->filter);
+		     panel->reverse, panel->case_sensitive,
+		     panel->exec_first, panel->filter);
     return panel;
 }
 
@@ -1084,7 +1090,7 @@ panel_reload (WPanel *panel)
     panel->count =
 	do_reload_dir (panel->cwd, &panel->dir, panel->sort_type,
 		       panel->count, panel->reverse, panel->case_sensitive,
-		       panel->filter);
+		       panel->exec_first, panel->filter);
 
     panel->dirty = 1;
     if (panel->selected >= panel->count)
@@ -1969,7 +1975,7 @@ do_enter_on_file_entry (file_entry *fe)
 	return 1;
 
     /* Check if the file is executable */
-    full_name = concat_dir_and_file (current_panel->cwd, fe->fname);
+    full_name = mhl_str_dir_plus_file (current_panel->cwd, fe->fname);
     if (!is_exe (fe->st.st_mode) || !if_link_is_exe (full_name, fe)) {
 	g_free (full_name);
 	return 0;
@@ -1988,7 +1994,7 @@ do_enter_on_file_entry (file_entry *fe)
 	char *tmp;
 	int ret;
 
-	tmp = concat_dir_and_file (vfs_get_current_dir (), fe->fname);
+	tmp = mhl_str_dir_plus_file (vfs_get_current_dir (), fe->fname);
 	ret = mc_setctl (tmp, VFS_SETCTL_RUN, NULL);
 	g_free (tmp);
 	/* We took action only if the dialog was shown or the execution
@@ -2025,10 +2031,10 @@ chdir_other_panel (WPanel *panel)
     }
 
     if (!S_ISDIR (panel->dir.list [panel->selected].st.st_mode)) {
-        new_dir = concat_dir_and_file (panel->cwd, "..");
+        new_dir = mhl_str_dir_plus_file (panel->cwd, "..");
 	sel_entry = strrchr(panel->cwd, PATH_SEP);
     } else
-        new_dir = concat_dir_and_file (panel->cwd, panel->dir.list [panel->selected].fname);
+        new_dir = mhl_str_dir_plus_file (panel->cwd, panel->dir.list [panel->selected].fname);
 
     change_panel ();
     do_cd (new_dir, cd_exact);
@@ -2094,7 +2100,7 @@ chdir_to_readlink (WPanel *panel)
 	if (*buffer == PATH_SEP)
 	    new_dir = g_strdup (buffer);
 	else
-	    new_dir = concat_dir_and_file (panel->cwd, buffer);
+	    new_dir = mhl_str_dir_plus_file (panel->cwd, buffer);
 
 	change_panel ();
 	do_cd (new_dir, cd_exact);
@@ -2460,7 +2466,8 @@ panel_re_sort (WPanel *panel)
 
     filename = g_strdup (selection (panel)->fname);
     unselect_item (panel);
-    do_sort (&panel->dir, panel->sort_type, panel->count-1, panel->reverse, panel->case_sensitive);
+    do_sort (&panel->dir, panel->sort_type, panel->count-1, panel->reverse,
+             panel->case_sensitive, panel->exec_first);
     panel->selected = -1;
     for (i = panel->count; i; i--){
 	if (!strcmp (panel->dir.list [i-1].fname, filename)){
