@@ -1,5 +1,6 @@
 /* Concurrent shell support for the Midnight Commander
-   Copyright (C) 1994, 1995 Dugan Porter
+   Copyright (C) 1994, 1995, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
+   2005, 2006, 2007 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of Version 2 of the GNU General Public
@@ -159,7 +160,7 @@ write_all (int fd, const void *buf, size_t count)
     ssize_t ret;
     ssize_t written = 0;
     while (count > 0) {
-	ret = write (fd, buf, count);
+	ret = write (fd, (const unsigned char *) buf + written, count);
 	if (ret < 0) {
 	    if (errno == EINTR) {
 		continue;
@@ -167,7 +168,6 @@ write_all (int fd, const void *buf, size_t count)
 		return written > 0 ? written : ret;
 	    }
 	}
-	buf += ret;
 	count -= ret;
 	written += ret;
     }
@@ -1110,6 +1110,7 @@ pty_open_slave (const char *pty_name)
 #endif				/* I_FIND && I_PUSH */
 #endif				/* __osf__ || __linux__ */
 
+    fcntl(pty_slave, F_SETFD, FD_CLOEXEC);
     return pty_slave;
 }
 
@@ -1163,6 +1164,7 @@ pty_open_slave (const char *pty_name)
     }
     if ((pty_slave = open (pty_name, O_RDWR)) == -1)
 	fprintf (stderr, "open (pty_name, O_RDWR): %s\r\n", pty_name);
+    fcntl(pty_slave, F_SETFD, FD_CLOEXEC);
     return pty_slave;
 }
 
