@@ -1,7 +1,8 @@
 /* Virtual File System: FISH implementation for transfering files over
    shell connections.
 
-   Copyright (C) 1998 The Free Software Foundation
+   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
+   2007 Free Software Foundation, Inc.
    
    Written by: 1998 Pavel Machek
    Spaces fix: 2000 Michal Svec
@@ -49,6 +50,9 @@
 #include "tcputil.h"
 #include "../src/unixcompat.h"
 #include "fish.h"
+#include "../mhl/memory.h"
+#include "../mhl/string.h"
+#include "../mhl/escape.h"
 
 int fish_directory_timeout = 900;
 
@@ -140,7 +144,7 @@ fish_command (struct vfs_class *me, struct vfs_s_super *super,
     enable_interrupt_key ();
 
     status = write (SUP.sockw, str, strlen (str));
-    g_free (str);
+    mhl_mem_free (str);
 
     disable_interrupt_key ();
     if (status < 0)
@@ -164,10 +168,10 @@ fish_free_archive (struct vfs_class *me, struct vfs_s_super *super)
 	close (SUP.sockr);
 	SUP.sockw = SUP.sockr = -1;
     }
-    g_free (SUP.host);
-    g_free (SUP.user);
-    g_free (SUP.cwdir);
-    g_free (SUP.password);
+    mhl_mem_free (SUP.host);
+    mhl_mem_free (SUP.user);
+    mhl_mem_free (SUP.cwdir);
+    mhl_mem_free (SUP.password);
 }
 
 static void
@@ -247,7 +251,7 @@ fish_open_archive_int (struct vfs_class *me, struct vfs_s_super *super)
 		p = g_strconcat (_(" fish: Password required for "),
 				 SUP.user, " ", (char *) NULL);
 		op = vfs_get_password (p);
-		g_free (p);
+		mhl_mem_free (p);
 		if (op == NULL)
 		    ERRNOR (EPERM, -1);
 		SUP.password = op;
@@ -310,7 +314,7 @@ fish_open_archive (struct vfs_class *me, struct vfs_s_super *super,
     p = vfs_split_url (strchr (op, ':') + 1, &host, &user, &flags,
 		       &password, 0, URL_NOSLASH);
 
-    g_free (p);
+    mhl_mem_free (p);
 
     SUP.host = host;
     SUP.user = user;
@@ -337,12 +341,12 @@ fish_archive_same (struct vfs_class *me, struct vfs_s_super *super,
     op = vfs_split_url (strchr (op, ':') + 1, &host, &user, &flags, 0, 0,
 			URL_NOSLASH);
 
-    g_free (op);
+    mhl_mem_free (op);
 
     flags = ((strcmp (host, SUP.host) == 0)
 	     && (strcmp (user, SUP.user) == 0) && (flags == SUP.flags));
-    g_free (host);
-    g_free (user);
+    mhl_mem_free (host);
+    mhl_mem_free (user);
 
     return flags;
 }
@@ -354,7 +358,18 @@ fish_dir_load(struct vfs_class *me, struct vfs_s_inode *dir, char *remote_path)
     char buffer[8192];
     struct vfs_s_entry *ent = NULL;
     FILE *logfile;
-    char *quoted_path;
+    SHELL_ESCAPED_STR quoted_path;
+    int reply_code;
+
+#if 0
+    /*
+     * Simple FISH debug interface :]
+     */
+    if (!(MEDATA->logfile))
+    {
+        MEDATA->logfile = fopen("/tmp/mc-FISH.sh", "w");
+    }
+#endif // 0
 
     logfile = MEDATA->logfile;
 
@@ -362,22 +377,84 @@ fish_dir_load(struct vfs_class *me, struct vfs_s_inode *dir, char *remote_path)
 
     gettimeofday(&dir->timestamp, NULL);
     dir->timestamp.tv_sec += fish_directory_timeout;
-    quoted_path = name_quote (remote_path, 0);
+    quoted_path = mhl_shell_escape_dup (remote_path);
     fish_command (me, super, NONE,
-	    "#LIST /%s\n"
-	    "ls -lLan /%s 2>/dev/null | grep '^[^cbt]' | (\n"
-	      "while read p l u g s m d y n; do\n"
-	        "echo \"P$p $u.$g\nS$s\nd$m $d $y\n:$n\n\"\n"
-	      "done\n"
-	    ")\n"
-	    "ls -lan /%s 2>/dev/null | grep '^[cb]' | (\n"
-	      "while read p l u g a i m d y n; do\n"
-	        "echo \"P$p $u.$g\nE$a$i\nd$m $d $y\n:$n\n\"\n"
-	      "done\n"
-	    ")\n"
-	    "echo '### 200'\n",
-	    remote_path, quoted_path, quoted_path);
-    g_free (quoted_path);
+		"#LIST /%s\n"
+		"if `perl -v > /dev/null 2>&1` ; then\n"
+		  "perl -e '\n"
+		   "use strict;\n"
+		   "use POSIX;\n"
+		   "use Fcntl;\n"
+		   "use POSIX \":fcntl_h\"; #S_ISLNK was here until 5.6\n"
+		   "import Fcntl \":mode\" unless defined &S_ISLNK; #and is now here\n"
+		   "my $dirname = $ARGV[0];\n"
+		   "if (opendir ( DIR, $dirname )) {\n"
+		   "while( (my $filename = readdir(DIR))){\n"
+		   "my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = lstat(\"$dirname/$filename\");\n"
+		   "my $mloctime= scalar localtime $mtime;\n"
+		   "my $shell_escape_regex= s/([;<>\\*\\|`&\\$!#\\(\\)\\[\\]\\{\\}:'\\''\"\\ \\\\])/\\\\$1/g;\n"
+		   "my $e_filename = $filename;\n"
+		   "$e_filename =~ $shell_escape_regex;\n"
+		   "if (S_ISLNK($mode) ) {\n"
+		   "my $linkname = readlink (\"$dirname/$filename\");\n"
+		   "$linkname =~ $shell_escape_regex;\n"
+		   "\n"
+			  "printf(\"R%%o %%o $uid.$gid\\n"
+				    "S$size\\n"
+				    "d$mloctime\\n"
+				    ":\\\"$e_filename\\\" -> \\\"$linkname\\\"\\n"
+				    "\\n\", S_IMODE($mode), S_IFMT($mode));\n"
+		   "} else {\n"
+			  "printf(\"R%%o %%o $uid.$gid\\n"
+				    "S$size\\n"
+				    "d$mloctime\\n"
+				    ":\\\"$e_filename\\\"\\n"
+				    "\\n\", S_IMODE($mode), S_IFMT($mode));\n"
+		   "}}\n"
+		   "printf(\"### 200\\n\");\n"
+		   "closedir(DIR);\n"
+		   "} else {\n"
+		    "printf(\"### 500\\n\");\n"
+		   "}\n"
+		   "exit 0\n"
+		   "' /%s ||\n" /* ARGV[0] - path to browse */
+		   "    echo '### 500'\n" /* do not hang if perl is to eval it */
+		"elif `ls -1 /%s >/dev/null 2>&1` ; then\n"
+		  "if `ls -Q /%s >/dev/null 2>&1`; then\n"
+			  "LSOPT=\"-Qlan\";\n"
+			  "ADD=0;\n"
+		  "else\n"
+			  "LSOPT=\"-lan\";\n"
+			  "ADD=1;\n"
+		  "fi\n"
+		  "ls $LSOPT \"/%s\" 2>/dev/null | grep '^[^cbt]' | (\n"
+			  "while read p l u g s m d y n; do\n"
+				  "if [ $ADD  = 0 ]; then\n"
+					  "echo \"P$p $u.$g\nS$s\nd$m $d $y\n:$n\n\"\n"
+				  "elif `sed --version >/dev/null 2>&1` ; then\n"
+					  "file=`echo $n | sed -e 's#^\\(.*\\) -> \\(.*\\)#\\1\" -> \"\\2#'`\n"
+					  "echo \"P$p $u $g\nS$s\nd$m $d $y\n:\"$file\"\n\"\n"
+				  "else\n"
+					  "echo \"P$p $u $g\nS$s\nd$m $d $y\n:\"$n\"\n\"\n"
+				  "fi\n"
+			  "done )\n"
+		  "ls $LSOPT /%s 2>/dev/null | grep '^[cb]' | (\n"
+			  "while read p l u g a i m d y n; do\n"
+				  "if [ $ADD = 0 ]; then\n"
+					  "echo \"P$p $u.$g\nE$a$i\nd$m $d $y\n:$n\n\"\n"
+				  "elif `sed --version >/dev/null 2>&1` ; then\n"
+					  "file=`echo $n | sed -e 's#^\\(.*\\) -> \\(.*\\)#\\1\" -> \"\\2#'`\n"
+					  "echo \"P$p $u $g\nS$s\nd$m $d $y\n:\"$file\"\n\"\n"
+				  "else\n"
+					  "echo \"P$p $u $g\nS$s\nd$m $d $y\n:\"$n\"\n\"\n"
+				  "fi\n"
+			  "done)\n"
+		  "echo '### 200'\n"
+	"else\n"
+		  "echo '### 500'\n"
+	"fi\n",
+	    quoted_path.s, quoted_path.s, quoted_path.s, quoted_path.s, quoted_path.s, quoted_path.s);
+    mhl_mem_free (quoted_path.s);
     ent = vfs_s_generate_entry(me, NULL, dir, 0);
     while (1) {
 	int res = vfs_s_get_line_interruptible (me, buffer, sizeof (buffer), SUP.sockr); 
@@ -405,11 +482,61 @@ fish_dir_load(struct vfs_class *me, struct vfs_s_inode *dir, char *remote_path)
 
 	switch(buffer[0]) {
 	case ':': {
-		      if (!strcmp(buffer+1, ".") || !strcmp(buffer+1, ".."))
-			  break;  /* We'll do . and .. ourself */
-		      ent->name = g_strdup(buffer+1); 
-		      break;
-	          }
+	    char *data_start = buffer+1;
+	    char *filename = data_start;
+	    char *linkname = data_start;
+	    char *filename_bound = filename + strlen(filename);
+	    char *linkname_bound = filename_bound;
+	    if (!strcmp(data_start, "\".\"") || !strcmp(data_start, "\"..\""))
+			break;  /* We'll do "." and ".." ourselves */
+
+		if (S_ISLNK(ST.st_mode)) {
+			// we expect: "escaped-name" -> "escaped-name"
+			//     -> cannot occur in filenames,
+			//     because it will be escaped to -\>
+
+			if (*filename == '"')
+				++filename;
+
+			linkname = strstr(filename, "\" -> \"");
+			if (!linkname)
+			{
+				/* broken client, or smth goes wrong */
+				linkname = filename_bound;
+				if (filename_bound > filename
+				    && *(filename_bound - 1) == '"')
+					--filename_bound; // skip trailing "
+			}
+			else
+			{
+				filename_bound = linkname;
+				linkname += 6; // strlen ("\" -> \"")
+				if (*(linkname_bound - 1) == '"')
+					--linkname_bound; // skip trailing "
+			}
+
+			ent->name = mhl_str_dup_range(filename, filename_bound);
+			mhl_shell_unescape_buf(ent->name);
+
+			ent->ino->linkname = mhl_str_dup_range(linkname, linkname_bound);
+			mhl_shell_unescape_buf(ent->ino->linkname);
+		} else {
+			// we expect: "escaped-name"
+			if (filename_bound - filename > 2)
+			{
+				// there is at least 2 "
+				// and we skip them
+				if (*filename == '"')
+					++filename;
+				if (*(filename_bound - 1) == '"')
+					--filename_bound;
+			}
+
+			ent->name = mhl_str_dup_range(filename, filename_bound);
+			mhl_shell_unescape_buf(ent->name);
+		}
+		break;
+	}
 	case 'S':
 #ifdef HAVE_ATOLL
 	    ST.st_size = (off_t) atoll (buffer+1);
@@ -419,11 +546,14 @@ fish_dir_load(struct vfs_class *me, struct vfs_s_inode *dir, char *remote_path)
 	    break;
 	case 'P': {
 	    size_t skipped;
-
-	    if (vfs_parse_filemode (buffer + 1, &skipped, &ST.st_mode)) {
-		if (S_ISLNK(ST.st_mode))
-		    ST.st_mode = 0;
-	    }
+	    vfs_parse_filemode (buffer + 1, &skipped, &ST.st_mode);
+	    break;
+	}
+	case 'R': {
+	    // raw filemode:
+	    // we expect: Roctal-filemode octal-filetype uid.gid
+	    size_t skipped;
+	    vfs_parse_raw_filemode (buffer + 1, &skipped, &ST.st_mode);
 	    break;
 	}
 	case 'd': {
@@ -449,23 +579,25 @@ fish_dir_load(struct vfs_class *me, struct vfs_s_inode *dir, char *remote_path)
 		      ST.st_rdev = makedev (maj, min);
 #endif
 	          }
-	case 'L': ent->ino->linkname = g_strdup(buffer+1);
-	          break;
 	}
     }
     
     vfs_s_free_entry (me, ent);
-    me->verrno = E_REMOTE;
-    if (fish_decode_reply(buffer+4, 0) == COMPLETE) {
-	g_free (SUP.cwdir);
+    reply_code = fish_decode_reply(buffer + 4, 0);
+    if (reply_code == COMPLETE) {
+	mhl_mem_free (SUP.cwdir);
 	SUP.cwdir = g_strdup (remote_path);
 	print_vfs_message (_("%s: done."), me->name);
 	return 0;
+    } else if (reply_code == ERROR) {
+	me->verrno = EACCES;
+    } else {
+	me->verrno = E_REMOTE;
     }
 
 error:
     print_vfs_message (_("%s: failure"), me->name);
-    return 1;
+    return -1;
 }
 
 static int
@@ -477,7 +609,7 @@ fish_file_store(struct vfs_class *me, struct vfs_s_fh *fh, char *name, char *loc
     struct stat s;
     int was_error = 0;
     int h;
-    char *quoted_name;
+    SHELL_ESCAPED_STR quoted_name;
 
     h = open (localname, O_RDONLY);
 
@@ -517,8 +649,8 @@ fish_file_store(struct vfs_class *me, struct vfs_s_fh *fh, char *name, char *loc
      *	algorithm for file appending case, therefore just "dd" is used for it.
      */
 
-    print_vfs_message(_("fish: store %s: sending command..."), name );
-    quoted_name = name_quote (name, 0);
+    quoted_name = mhl_shell_escape_dup(name);
+    print_vfs_message(_("fish: store %s: sending command..."), quoted_name.s );
 
     /* FIXME: File size is limited to ULONG_MAX */
     if (!fh->u.fish.append)
@@ -530,7 +662,7 @@ fish_file_store(struct vfs_class *me, struct vfs_s_fh *fh, char *name, char *loc
 		 "(\n"
 		   "head -c %lu -q - || echo DD >&3\n"
 		 ") 2>/dev/null | (\n"
-		   "cat > \"$file\"\n"
+		   "cat > $file\n"
 		   "cat > /dev/null\n"
 		 ")`; [ \"$res\" = DD ] && {\n"
 			"> \"$file\"\n"
@@ -542,8 +674,8 @@ fish_file_store(struct vfs_class *me, struct vfs_s_fh *fh, char *name, char *loc
 			"    rest=`expr $rest - $n`\n"
 			"done\n"
 		 "}; echo '### 200'\n",
-		 (unsigned long) s.st_size, name,
-		 quoted_name, (unsigned long) s.st_size,
+		 (unsigned long) s.st_size, quoted_name.s,
+		 quoted_name.s, (unsigned long) s.st_size,
 		 (unsigned long) s.st_size);
     else
 	n = fish_command (me, super, WAIT_REPLY,
@@ -555,14 +687,13 @@ fish_file_store(struct vfs_class *me, struct vfs_s_fh *fh, char *name, char *loc
 			"while [ $rest -gt 0 ]\n"
 			"do\n"
 			"    cnt=`expr \\( $rest + 255 \\) / 256`\n"
-			"    n=`dd bs=256 count=$cnt | tee -a \"$file\" | wc -c`\n"
+			"    n=`dd bs=256 count=$cnt | tee -a $file | wc -c`\n"
 			"    rest=`expr $rest - $n`\n"
 			"done\n"
 		 "}; echo '### 200'\n",
-		 (unsigned long) s.st_size, name,
-		 quoted_name, (unsigned long) s.st_size);
+		 (unsigned long) s.st_size, quoted_name.s,
+		 quoted_name.s, (unsigned long) s.st_size);
 
-    g_free (quoted_name);
     if (n != PRELIM) {
 	close (h);
         ERRNOR(E_REMOTE, -1);
@@ -595,12 +726,14 @@ fish_file_store(struct vfs_class *me, struct vfs_s_fh *fh, char *name, char *loc
 			  (unsigned long) s.st_size);
     }
     close(h);
+    mhl_mem_free(quoted_name.s);
     if ((fish_get_reply (me, SUP.sockr, NULL, 0) != COMPLETE) || was_error)
         ERRNOR (E_REMOTE, -1);
     return 0;
 error_return:
     close(h);
     fish_get_reply(me, SUP.sockr, NULL, 0);
+    mhl_mem_free(quoted_name.s);
     return -1;
 }
 
@@ -608,35 +741,47 @@ static int
 fish_linear_start (struct vfs_class *me, struct vfs_s_fh *fh, off_t offset)
 {
     char *name;
-    char *quoted_name;
+    SHELL_ESCAPED_STR quoted_name;
     if (offset)
         ERRNOR (E_NOTSUPP, 0);
     name = vfs_s_fullpath (me, fh->ino);
     if (!name)
 	return 0;
-    quoted_name = name_quote (name, 0);
-    g_free (name);
-    name = quoted_name;
+    quoted_name = mhl_shell_escape_dup(name);
     fh->u.fish.append = 0;
+
+    /*
+     * Check whether the remote file is readable by using `dd' to copy 
+     * a single byte from the remote file to /dev/null. If `dd' completes
+     * with exit status of 0 use `cat' to send the file contents to the
+     * standard output (i.e. over the network).
+     */
     offset = fish_command (me, FH_SUPER, WANT_STRING,
 		"#RETR /%s\n"
+		"if dd if=/%s of=/dev/null bs=1 count=1 2>/dev/null ;\n"
+		"then\n"
 		"ls -ln /%s 2>/dev/null | (\n"
 		  "read p l u g s r\n"
-		  "echo \"$s\"\n"
+		  "echo $s\n"
 		")\n"
 		"echo '### 100'\n"
 		"cat /%s\n"
-		"echo '### 200'\n", 
-		name, name, name );
-    g_free (name);
+		"echo '### 200'\n"
+		"else\n"
+		"echo '### 500'\n"
+		"fi\n",
+		quoted_name.s, quoted_name.s, quoted_name.s, quoted_name.s );
+    mhl_mem_free (quoted_name.s);
     if (offset != PRELIM) ERRNOR (E_REMOTE, 0);
     fh->linear = LS_LINEAR_OPEN;
     fh->u.fish.got = 0;
-#if defined(_FILE_OFFSET_BITS) && _FILE_OFFSET_BITS == 64
-    if (sscanf( reply_str, "%llu", &fh->u.fish.total )!=1)
+    errno = 0;
+#if SIZEOF_OFF_T == SIZEOF_LONG
+    fh->u.fish.total = strtol (reply_str, NULL, 10);
 #else
-    if (sscanf( reply_str, "%u", &fh->u.fish.total )!=1)
+    fh->u.fish.total = strtoll (reply_str, NULL, 10);
 #endif
+    if (errno != 0)
 	ERRNOR (E_REMOTE, 0);
     return 1;
 }
@@ -735,17 +880,18 @@ fish_send_command(struct vfs_class *me, struct vfs_s_super *super, const char *c
 #define PREFIX \
     char buf[BUF_LARGE]; \
     const char *crpath; \
-    char *rpath, *mpath = g_strdup (path); \
+    char *mpath = mhl_str_dup (path); \
+    SHELL_ESCAPED_STR rpath; \
     struct vfs_s_super *super; \
     if (!(crpath = vfs_s_get_path_mangle (me, mpath, &super, 0))) { \
-	g_free (mpath); \
+	mhl_mem_free (mpath); \
 	return -1; \
     } \
-    rpath = name_quote (crpath, 0); \
-    g_free (mpath);
+    rpath = mhl_shell_escape_dup(crpath); \
+    mhl_mem_free (mpath);
 
 #define POSTFIX(flags) \
-    g_free (rpath); \
+    mhl_mem_free (rpath.s); \
     return fish_send_command(me, super, buf, flags);
 
 static int
@@ -753,58 +899,57 @@ fish_chmod (struct vfs_class *me, const char *path, int mode)
 {
     PREFIX
     g_snprintf(buf, sizeof(buf), "#CHMOD %4.4o /%s\n"
-				 "chmod %4.4o \"/%s\" 2>/dev/null\n"
+				 "chmod %4.4o /%s 2>/dev/null\n"
 				 "echo '### 000'\n", 
-	    mode & 07777, rpath,
-	    mode & 07777, rpath);
+	    mode & 07777, rpath.s,
+	    mode & 07777, rpath.s);
     POSTFIX(OPT_FLUSH);
 }
 
-#define FISH_OP(name, chk, string) \
+#define FISH_OP(name, string) \
 static int fish_##name (struct vfs_class *me, const char *path1, const char *path2) \
 { \
     char buf[BUF_LARGE]; \
     const char *crpath1, *crpath2; \
-    char *rpath1, *rpath2, *mpath1, *mpath2; \
+    char *mpath1, *mpath2; \
     struct vfs_s_super *super1, *super2; \
     if (!(crpath1 = vfs_s_get_path_mangle (me, mpath1 = g_strdup(path1), &super1, 0))) { \
-	g_free (mpath1); \
+	mhl_mem_free (mpath1); \
 	return -1; \
     } \
     if (!(crpath2 = vfs_s_get_path_mangle (me, mpath2 = g_strdup(path2), &super2, 0))) { \
-	g_free (mpath1); \
-	g_free (mpath2); \
+	mhl_mem_free (mpath1); \
+	mhl_mem_free (mpath2); \
 	return -1; \
     } \
-    rpath1 = name_quote (crpath1, 0); \
-    g_free (mpath1); \
-    rpath2 = name_quote (crpath2, 0); \
-    g_free (mpath2); \
-    g_snprintf(buf, sizeof(buf), string "\n", rpath1, rpath2, rpath1, rpath2); \
-    g_free (rpath1); \
-    g_free (rpath2); \
+    SHELL_ESCAPED_STR rpath1 = mhl_shell_escape_dup (crpath1); \
+    mhl_mem_free (mpath1); \
+    SHELL_ESCAPED_STR rpath2 = mhl_shell_escape_dup (crpath2); \
+    mhl_mem_free (mpath2); \
+    g_snprintf(buf, sizeof(buf), string "\n", rpath1.s, rpath2.s, rpath1.s, rpath2.s); \
+    mhl_mem_free (rpath1.s); \
+    mhl_mem_free (rpath2.s); \
     return fish_send_command(me, super2, buf, OPT_FLUSH); \
 }
 
-#define XTEST if (bucket1 != bucket2) { ERRNOR (EXDEV, -1); }
-FISH_OP(rename, XTEST, "#RENAME /%s /%s\n"
-		       "mv /%s /%s 2>/dev/null\n"
-		       "echo '### 000'" )
-FISH_OP(link,   XTEST, "#LINK /%s /%s\n"
-		       "ln /%s /%s 2>/dev/null\n"
-		       "echo '### 000'" )
+FISH_OP(rename, "#RENAME /%s /%s\n"
+		"mv /%s /%s 2>/dev/null\n"
+		"echo '### 000'" )
+FISH_OP(link,   "#LINK /%s /%s\n"
+		"ln /%s /%s 2>/dev/null\n"
+		"echo '### 000'" )
 
 static int fish_symlink (struct vfs_class *me, const char *setto, const char *path)
 {
-    char *qsetto;
+    SHELL_ESCAPED_STR qsetto;
     PREFIX
-    qsetto = name_quote (setto, 0);
+    qsetto = mhl_shell_escape_dup (setto);
     g_snprintf(buf, sizeof(buf),
             "#SYMLINK %s /%s\n"
 	    "ln -s %s /%s 2>/dev/null\n"
 	    "echo '### 000'\n",
-	    qsetto, rpath, qsetto, rpath);
-    g_free (qsetto);
+	    qsetto.s, rpath.s, qsetto.s, rpath.s);
+    mhl_mem_free (qsetto.s);
     POSTFIX(OPT_FLUSH);
 }
 
@@ -826,19 +971,19 @@ fish_chown (struct vfs_class *me, const char *path, int owner, int group)
     {
 	PREFIX
 	g_snprintf (buf, sizeof(buf),
-    	    "#CHOWN /%s /%s\n"
+    	    "#CHOWN %s /%s\n"
 	    "chown %s /%s 2>/dev/null\n"
 	    "echo '### 000'\n", 
-	    sowner, rpath,
-	    sowner, rpath);
+	    sowner, rpath.s,
+	    sowner, rpath.s);
 	fish_send_command (me, super, buf, OPT_FLUSH); 
 	/* FIXME: what should we report if chgrp succeeds but chown fails? */
 	g_snprintf (buf, sizeof(buf),
-            "#CHGRP /%s /%s\n"
-	    "chgrp %s /%s 2>/dev/null\n"
+            "#CHGRP /%s \"/%s\"\n"
+	    "chgrp %s \"/%s\" 2>/dev/null\n"
 	    "echo '### 000'\n", 
-	    sgroup, rpath,
-	    sgroup, rpath);
+	    sgroup, rpath.s,
+	    sgroup, rpath.s);
 	/* fish_send_command(me, super, buf, OPT_FLUSH); */
 	POSTFIX (OPT_FLUSH)
     }
@@ -851,7 +996,7 @@ static int fish_unlink (struct vfs_class *me, const char *path)
             "#DELE /%s\n"
 	    "rm -f /%s 2>/dev/null\n"
 	    "echo '### 000'\n",
-	    rpath, rpath);
+	    rpath.s, rpath.s);
     POSTFIX(OPT_FLUSH);
 }
 
@@ -865,7 +1010,7 @@ static int fish_mkdir (struct vfs_class *me, const char *path, mode_t mode)
             "#MKD /%s\n"
 	    "mkdir /%s 2>/dev/null\n"
 	    "echo '### 000'\n",
-	    rpath, rpath);
+	    rpath.s, rpath.s);
     POSTFIX(OPT_FLUSH);
 }
 
@@ -876,7 +1021,7 @@ static int fish_rmdir (struct vfs_class *me, const char *path)
             "#RMD /%s\n"
 	    "rmdir /%s 2>/dev/null\n"
 	    "echo '### 000'\n",
-	    rpath, rpath);
+	    rpath.s, rpath.s);
     POSTFIX(OPT_FLUSH);
 }
 
@@ -934,7 +1079,7 @@ fish_fill_names (struct vfs_class *me, fill_names_f func)
 	name = g_strconcat ("/#sh:", SUP.user, "@", SUP.host, flags,
 			    "/", SUP.cwdir, (char *) NULL);
 	(*func)(name);
-	g_free (name);
+	mhl_mem_free (name);
 	super = super->next;
     }
 }
