@@ -1,7 +1,22 @@
 /*
  * Single File fileSystem
  *
- * Copyright 1998 Pavel Machek, distribute under GPL
+ * Copyright 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007
+ * Free Software Foundation, Inc.
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  * This defines whole class of filesystems which contain single file
  * inside. It is somehow similar to extfs, except that extfs makes
@@ -19,10 +34,18 @@
 #include <stdio.h>
 #include <string.h>
 
-#include "utilvfs.h"
+#include <mhl/string.h>
 
+#include "../src/global.h"
+#include "../src/tty.h"		/* enable/disable interrupt key */
+#include "../src/wtools.h"	/* message() */
+#include "../src/main.h"	/* print_vfs_message */
+#include "utilvfs.h"
 #include "vfs.h"
+#include "vfs-impl.h"
+#include "gc.h"		/* vfs_stamp_create */
 #include "local.h"
+#include "../src/execute.h"	/* EXECUTE_AS_SHELL */
 
 struct cachedfile {
     char *name, *cache;
@@ -30,6 +53,7 @@ struct cachedfile {
 };
 
 static struct cachedfile *head;
+static struct vfs_class vfs_sfs_ops;
 
 #define MAXFS 32
 static int sfs_no = 0;
@@ -41,46 +65,77 @@ static int sfs_flags[ MAXFS ];
 #define F_NOLOCALCOPY 4
 #define F_FULLMATCH 8
 
-static int uptodate (char *name, char *cache)
-{
-    return 1;
-}
-
-static int vfmake (vfs *me, char *name, char *cache)
+static int
+sfs_vfmake (struct vfs_class *me, const char *name, char *cache)
 {
     char *inpath, *op;
     int w;
-    char pad [10240];
+    char pad[10240];
     char *s, *t = pad;
     int was_percent = 0;
+    char *pname;	/* name of parent archive */
+    char *pqname;	/* name of parent archive, quoted */
 
-    vfs_split (name, &inpath, &op);
+    pname = g_strdup (name);
+    vfs_split (pname, &inpath, &op);
     if ((w = (*me->which) (me, op)) == -1)
-        vfs_die ("This cannot happen... Hopefully.\n");
+	vfs_die ("This cannot happen... Hopefully.\n");
+
+    if (!(sfs_flags[w] & F_1) && strcmp (pname, "/")) {
+	g_free (pname);
+	return -1;
+    }
 
-    if ((sfs_flags[w] & F_1) || (!strcmp (name, "/"))) ; else return -1;
     /*    if ((sfs_flags[w] & F_2) || (!inpath) || (!*inpath)); else return -1; */
     if (!(sfs_flags[w] & F_NOLOCALCOPY)) {
-        s = mc_getlocalcopy (name);
-	if (!s)
+	s = mc_getlocalcopy (pname);
+	if (!s) {
+	    g_free (pname);
 	    return -1;
-        name = name_quote (s, 0);
+	}
+	pqname = name_quote (s, 0);
 	g_free (s);
-    } else 
-        name = name_quote (name, 0);
-#define COPY_CHAR if (t-pad>sizeof(pad)) { g_free (name); return -1; } else *t++ = *s;
-#define COPY_STRING(a) if ((t-pad)+strlen(a)>sizeof(pad)) { g_free (name); return -1; } else { strcpy (t, a); t+= strlen(a); }
-    for (s = sfs_command[w]; *s; s++) {
-        if (was_percent) {
+    } else {
+	pqname = name_quote (pname, 0);
+    }
+    g_free (pname);
 
-	    char *ptr = NULL;
+#define COPY_CHAR \
+    if ((size_t) (t-pad) > sizeof(pad)) { \
+	g_free (pqname); \
+	return -1; \
+    } \
+    else \
+	*t++ = *s;
+
+#define COPY_STRING(a) \
+    if ((t-pad)+strlen(a)>sizeof(pad)) { \
+	g_free (pqname); \
+	return -1; \
+    } else { \
+	strcpy (t, a); \
+	t+= strlen(a); \
+    }
+
+    for (s = sfs_command[w]; *s; s++) {
+	if (was_percent) {
+
+	    const char *ptr = NULL;
 	    was_percent = 0;
 
 	    switch (*s) {
-	    case '1': ptr = name; break;
-	    case '2': ptr = op + strlen (sfs_prefix[w]); break;
-	    case '3': ptr = cache; break;
-	    case '%': COPY_CHAR; continue; 
+	    case '1':
+		ptr = pqname;
+		break;
+	    case '2':
+		ptr = op + strlen (sfs_prefix[w]);
+		break;
+	    case '3':
+		ptr = cache;
+		break;
+	    case '%':
+		COPY_CHAR;
+		continue;
 	    }
 	    COPY_STRING (ptr);
 	} else {
@@ -90,8 +145,8 @@ static int vfmake (vfs *me, char *name, char *cache)
 		COPY_CHAR;
 	}
     }
-    g_free (name);
 
+    g_free (pqname);
     open_error_pipe ();
     if (my_system (EXECUTE_AS_SHELL, "/bin/sh", pad)) {
 	close_error_pipe (1, NULL);
@@ -99,28 +154,25 @@ static int vfmake (vfs *me, char *name, char *cache)
     }
 
     close_error_pipe (0, NULL);
-    return 0; /* OK */
+    return 0;			/* OK */
 }
 
-static char *
-redirect (vfs *me, char *name)
+static const char *
+sfs_redirect (struct vfs_class *me, const char *name)
 {
     struct cachedfile *cur = head;
     char *cache;
     int handle;
 
     while (cur) {
-	/* FIXME: when not uptodate, we might want to kill cache
-	 * file immediately, not to wait until timeout. */
-	if ((!strcmp (name, cur->name))
-	    && (uptodate (cur->name, cur->cache))) {
+	if (!strcmp (name, cur->name)) {
 	    vfs_stamp (&vfs_sfs_ops, cur);
 	    return cur->cache;
 	}
 	cur = cur->next;
     }
 
-    handle = mc_mkstemps (&cache, "sfs", NULL);
+    handle = vfs_mkstemps (&cache, "sfs", name);
 
     if (handle == -1) {
 	return "/SOMEONE_PLAYING_DIRTY_TMP_TRICKS_ON_US";
@@ -128,15 +180,14 @@ redirect (vfs *me, char *name)
 
     close (handle);
 
-    if (!vfmake (me, name, cache)) {
+    if (!sfs_vfmake (me, name, cache)) {
 	cur = g_new (struct cachedfile, 1);
 	cur->name = g_strdup (name);
 	cur->cache = cache;
 	cur->next = head;
 	head = cur;
 
-	vfs_add_noncurrent_stamps (&vfs_sfs_ops, (vfsid) head, NULL);
-	vfs_rm_parents (NULL);
+	vfs_stamp_create (&vfs_sfs_ops, head);
 
 	return cache;
     }
@@ -147,31 +198,31 @@ redirect (vfs *me, char *name)
 }
 
 static void *
-sfs_open (vfs *me, char *path, int flags, int mode)
+sfs_open (struct vfs_class *me, const char *path, int flags, int mode)
 {
     int *sfs_info;
     int fd;
 
-    path = redirect (me, path);
-    fd = open (path, NO_LINEAR(flags), mode);
+    path = sfs_redirect (me, path);
+    fd = open (path, NO_LINEAR (flags), mode);
     if (fd == -1)
 	return 0;
 
     sfs_info = g_new (int, 1);
     *sfs_info = fd;
-    
+
     return sfs_info;
 }
 
-static int sfs_stat (vfs *me, char *path, struct stat *buf)
+static int sfs_stat (struct vfs_class *me, const char *path, struct stat *buf)
 {
-    path = redirect (me, path);
+    path = sfs_redirect (me, path);
     return stat (path, buf);
 }
 
-static int sfs_lstat (vfs *me, char *path, struct stat *buf)
+static int sfs_lstat (struct vfs_class *me, const char *path, struct stat *buf)
 {
-    path = redirect (me, path);
+    path = sfs_redirect (me, path);
 #ifndef HAVE_STATLSTAT
     return lstat (path, buf);
 #else
@@ -179,69 +230,44 @@ static int sfs_lstat (vfs *me, char *path, struct stat *buf)
 #endif
 }
 
-static int sfs_chmod (vfs *me, char *path, int mode)
+static int sfs_chmod (struct vfs_class *me, const char *path, int mode)
 {
-    path = redirect (me, path);
+    path = sfs_redirect (me, path);
     return chmod (path, mode);
 }
 
-static int sfs_chown (vfs *me, char *path, int owner, int group)
+static int sfs_chown (struct vfs_class *me, const char *path, int owner, int group)
 {
-    path = redirect (me, path);
+    path = sfs_redirect (me, path);
     return chown (path, owner, group);
 }
 
-static int sfs_utime (vfs *me, char *path, struct utimbuf *times)
+static int sfs_utime (struct vfs_class *me, const char *path, struct utimbuf *times)
 {
-    path = redirect (me, path);
+    path = sfs_redirect (me, path);
     return utime (path, times);
 }
 
-static int sfs_readlink (vfs *me, char *path, char *buf, int size)
+static int
+sfs_readlink (struct vfs_class *me, const char *path, char *buf, size_t size)
 {
-    path = redirect (me, path);
+    path = sfs_redirect (me, path);
     return readlink (path, buf, size);
 }
 
 static vfsid
-sfs_getid (vfs *me, char *path, struct vfs_stamping **parent)
-{				/* FIXME: what should I do? */
-    vfs *v;
-    vfsid id;
-    struct vfs_stamping *par;
+sfs_getid (struct vfs_class *me, const char *path)
+{
     struct cachedfile *cur = head;
 
+    (void) me;
+
     while (cur) {
 	if (!strcmp (path, cur->name))
 	    break;
 	cur = cur->next;
     }
 
-    *parent = NULL;
-
-    if (!cur)
-	return (vfsid) (-1);
-
-    {
-	char *path2 = g_strdup (path);
-
-	/* Strip suffix which led to this being sfs */
-	v = vfs_split (path2, NULL, NULL);
-
-	/* ... and learn whoever was the parent system */
-	v = vfs_split (path2, NULL, NULL);
-
-	id = (*v->getid) (v, path2, &par);
-	g_free (path2);
-    }
-
-    if (id != (vfsid) - 1) {
-	*parent = g_new (struct vfs_stamping, 1);
-	(*parent)->v = v;
-	(*parent)->id = id;
-	(*parent)->parent = par;
-	(*parent)->next = NULL;
-    }
     return (vfsid) cur;
 }
 
@@ -266,10 +292,12 @@ static void sfs_free (vfsid id)
     g_free (cur);
 }
 
-static void sfs_fill_names (vfs *me, void (*func)(char *))
+static void sfs_fill_names (struct vfs_class *me, fill_names_f func)
 {
     struct cachedfile *cur = head;
 
+    (void) me;
+
     while (cur){
 	(*func)(cur->name);
 	cur = cur->next;
@@ -280,27 +308,37 @@ static int sfs_nothingisopen (vfsid id)
 {
     /* FIXME: Investigate whether have to guard this like in
        the other VFSs (see fd_usage in extfs) -- Norbert */
+    (void) id;
     return 1;
 }
 
-static char *sfs_getlocalcopy (vfs *me, char *path)
+static char *
+sfs_getlocalcopy (struct vfs_class *me, const char *path)
 {
-    path = redirect (me, path);
+    path = sfs_redirect (me, path);
     return g_strdup (path);
 }
 
-static int sfs_ungetlocalcopy (vfs *me, char *path, char *local, int has_changed)
+static int
+sfs_ungetlocalcopy (struct vfs_class *me, const char *path,
+		    const char *local, int has_changed)
 {
-    g_free(local);
+    (void) me;
+    (void) path;
+    (void) local;
+    (void) has_changed;
     return 0;
 }
 
-static int sfs_init (vfs *me)
+static int sfs_init (struct vfs_class *me)
 {
     char *mc_sfsini;
     FILE *cfg;
+    char key[256];
 
-    mc_sfsini = concat_dir_and_file (mc_home, "extfs" PATH_SEP_STR "sfs.ini");
+    (void) me;
+
+    mc_sfsini = mhl_str_dir_plus_file (mc_home, "extfs" PATH_SEP_STR "sfs.ini");
     cfg = fopen (mc_sfsini, "r");
 
     if (!cfg){
@@ -311,14 +349,10 @@ static int sfs_init (vfs *me)
     g_free (mc_sfsini);
 
     sfs_no = 0;
-    while (sfs_no < MAXFS){
-	char key[256];
+    while (sfs_no < MAXFS && fgets (key, sizeof (key), cfg)) {
 	char *c, *semi = NULL, flags = 0;
 
-        if (!fgets (key, sizeof (key), cfg))
-	    break;
-
-	if (*key == '#')
+	if (*key == '#' || *key == '\n')
 	    continue;
 
 	for (c = key; *c; c++)
@@ -332,13 +366,14 @@ static int sfs_init (vfs *me)
 	    }
 
 	if (!semi){
+	invalid_line:
 	    fprintf (stderr, _("Warning: Invalid line in %s:\n%s\n"),
 		     "sfs.ini", key);
 	    continue;
 	}
 
 	c = semi + 1;
-	while ((*c != ' ') && (*c != '\t')) {
+	while (*c && (*c != ' ') && (*c != '\t')) {
 	    switch (*c) {
 	    case '1': flags |= F_1; break;
 	    case '2': flags |= F_2; break;
@@ -349,6 +384,8 @@ static int sfs_init (vfs *me)
 	    }	    
 	    c++;
 	}
+	if (!*c)
+	    goto invalid_line;
 	c++;
 	*(semi+1) = 0;
 	if ((semi = strchr (c, '\n')))
@@ -364,10 +401,12 @@ static int sfs_init (vfs *me)
 }
 
 static void
-sfs_done (vfs *me)
+sfs_done (struct vfs_class *me)
 {
     int i;
 
+    (void) me;
+
     for (i = 0; i < sfs_no; i++){
         g_free (sfs_prefix [i]);
 	g_free (sfs_command [i]);
@@ -377,10 +416,12 @@ sfs_done (vfs *me)
 }
 
 static int
-sfs_which (vfs *me, char *path)
+sfs_which (struct vfs_class *me, const char *path)
 {
     int i;
 
+    (void) me;
+
     for (i = 0; i < sfs_no; i++)
         if (sfs_flags [i] & F_FULLMATCH) {
 	    if (!strcmp (path, sfs_prefix [i]))
@@ -392,63 +433,30 @@ sfs_which (vfs *me, char *path)
     return -1;
 }
 
-vfs vfs_sfs_ops = {
-    NULL,	/* This is place of next pointer */
-    "sfs",
-    F_EXEC,	/* flags */
-    NULL,	/* prefix */
-    NULL,	/* data */
-    0,		/* errno */
-    sfs_init,
-    sfs_done,
-    sfs_fill_names,
-    sfs_which,
-
-    sfs_open,
-    local_close,
-    local_read,
-    NULL,
-    
-    NULL,
-    NULL,
-    NULL,
-    NULL,
-    NULL,
-
-    sfs_stat,
-    sfs_lstat,
-    local_fstat,
-
-    sfs_chmod,
-    sfs_chown,
-    sfs_utime,
-
-    sfs_readlink,
-    NULL,
-    NULL,
-    NULL,
-
-    NULL,
-    NULL,
-    local_errno,
-    local_lseek,
-    NULL,
-    
-    sfs_getid,
-    sfs_nothingisopen,
-    sfs_free,
-    
-    sfs_getlocalcopy,
-    sfs_ungetlocalcopy,
-    
-    NULL,
-    NULL,
-    NULL,
-    NULL
-
-#ifdef HAVE_MMAP
-    ,local_mmap,
-    local_munmap
-#endif
-};
-
+void
+init_sfs (void)
+{
+    vfs_sfs_ops.name = "sfs";
+    vfs_sfs_ops.init = sfs_init;
+    vfs_sfs_ops.done = sfs_done;
+    vfs_sfs_ops.fill_names = sfs_fill_names;
+    vfs_sfs_ops.which = sfs_which;
+    vfs_sfs_ops.open = sfs_open;
+    vfs_sfs_ops.close = local_close;
+    vfs_sfs_ops.read = local_read;
+    vfs_sfs_ops.stat = sfs_stat;
+    vfs_sfs_ops.lstat = sfs_lstat;
+    vfs_sfs_ops.fstat = local_fstat;
+    vfs_sfs_ops.chmod = sfs_chmod;
+    vfs_sfs_ops.chown = sfs_chown;
+    vfs_sfs_ops.utime = sfs_utime;
+    vfs_sfs_ops.readlink = sfs_readlink;
+    vfs_sfs_ops.ferrno = local_errno;
+    vfs_sfs_ops.lseek = local_lseek;
+    vfs_sfs_ops.getid = sfs_getid;
+    vfs_sfs_ops.nothingisopen = sfs_nothingisopen;
+    vfs_sfs_ops.free = sfs_free;
+    vfs_sfs_ops.getlocalcopy = sfs_getlocalcopy;
+    vfs_sfs_ops.ungetlocalcopy = sfs_ungetlocalcopy;
+    vfs_register_class (&vfs_sfs_ops);
+}
